import numpy as np
import os
import pycuda.driver as cuda
import pycuda.autoinit
import tensorrt as trt


def inference_with_tensorRT(engine_path: str, input_data: np.ndarray):

    """快速使用tensorRT引擎进行推理,用于测试可行性和准确性

    Args:
        engine_path: str类型,tensorRT引擎文件路径
        input_data= np.ndarray,输入的数据

    Returns:
        推理后的结果

    """

    # 创建tensorRT的记录器
    TRT_LOGGER = trt.Logger(trt.Logger.INFO)
    # 获取引擎和语境
    assert os.path.exists(engine_path)
    f = open(engine_path, "rb")
    runtime = trt.Runtime(TRT_LOGGER)
    engine = runtime.deserialize_cuda_engine(f.read())
    context = engine.create_execution_context()
    # 输出的缓冲区和计算区
    outputs_buffer = []
    outputs_memory = []
    bindings = []
    # 申请空间
    for location, binding in enumerate(engine):
        # 获取每个binding的信息并申请空间
        size = trt.volume(engine.get_binding_shape(binding))
        dtype = trt.nptype(engine.get_binding_dtype(binding))

        # 打印binding的信息
        # print("binding name:", engine.get_binding_name(location))
        # print("binding size:", size)
        # print("binding size:", dtype)

        buffer = cuda.pagelocked_empty(size, dtype)
        memory = cuda.mem_alloc(buffer.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(memory))

        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            input_memory = memory
        else:
            outputs_buffer.append(buffer)
            outputs_memory.append(memory)

    stream = cuda.Stream()

    # 将输入数据从缓存区读入内存区
    input_data = np.ascontiguousarray(input_data)
    cuda.memcpy_htod_async(input_memory, input_data, stream)

    # 推理
    context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)

    # 将输出数据从内存区读出缓存区
    for output_buffer, output_memory in zip(outputs_buffer, outputs_memory):
        cuda.memcpy_dtoh_async(output_buffer, output_memory, stream)

    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    f.close()
    return outputs_buffer


def onnx2trt(max_workspace_size=28, target_engine_path="./test.engine", onnx_file_path="./test.onnx"):

    """将onnx文件转换为tensorRT引擎文件

    Args:
        max_workspace_size: int类型,二进制式工作空间大小
        target_engine_path= str类型,输出的引擎文件路径
        onnx_file_path= str类型,输入的onnx文件路径

    Returns:
        无

    """
    logger = trt.Logger(trt.Logger.WARNING)
    builder = trt.Builder(logger)
    network = builder.create_network(1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH))
    parser = trt.OnnxParser(network, logger)
    # 测试是否成功
    success = parser.parse_from_file(onnx_file_path)
    for idx in range(parser.num_errors):
        print(parser.get_error(idx))
    config = builder.create_builder_config()
    config.max_workspace_size = 1 << max_workspace_size  # 1 MiB
    # tensorRT8的教程格式
    # config.set_memory_pool_limit(trt.MemoryPoolType.WORKSPACE, 1 << 28) # 1 MiB
    serialized_engine = builder.build_serialized_network(network, config)
    with open(target_engine_path, "wb") as f:
        f.write(serialized_engine)